/***************************************************************************
                          kttt3d.cpp  -  description
                             -------------------
    begin                :    8 22:37:07 MSD 2000
    copyright            : (C) 2000 by Andrei Plekhanov
    email                : andyp@newtech.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kttt3d.h"
#include <qkeycode.h>
#include <qmessagebox.h>
#include <qpainter.h>
#include <qevent.h>
#include "ttt3dgame.h"

const int SB_MESSAGE = 1;

Kttt3d::Kttt3d( void ) : KTopLevelWidget()
{
  locale = kapp->getLocale();

  setCaption( locale->translate("Kttt3d") );

  QPopupMenu *file = new QPopupMenu();
  CHECK_PTR( file );
  file->insertItem( i18n("&New game"),  this, SLOT(NewGame()), CTRL+Key_N );
  file->insertSeparator();
  file->insertItem( locale->translate("E&xit"),  qApp, SLOT(quit()), CTRL+Key_X );

  edit = new QPopupMenu();
  CHECK_PTR( edit );
  undoID = edit->insertItem( locale->translate("&Undo"), this, SLOT(undo()) );
  RevertID = edit->insertItem( locale->translate("&Revert"), this, SLOT(revert()) );
  edit->setItemEnabled( undoID, FALSE );

  skill = new QPopupMenu();
  CHECK_PTR( skill );
  skill->setCheckable( TRUE );
  SkillIDArr[ 0 ] = skill->insertItem( locale->translate("&Easy"),  this, SLOT(easy()), CTRL+Key_E );
  SkillIDArr[ 1 ] = skill->insertItem( locale->translate("&Normal"),  this, SLOT(normal()), CTRL+Key_N );
  SkillIDArr[ 2 ] = skill->insertItem( locale->translate("&Difficult"),  this, SLOT(difficult()), CTRL+Key_D );
  Skill = Ttt3dGame::BEGINNER;
  skill->setItemChecked( SkillIDArr[ 0 ], TRUE );

  QPopupMenu *help = new QPopupMenu;
  CHECK_PTR( help );
  help->insertItem( locale->translate("&About"), this, SLOT(about()), CTRL+Key_H );
  help->insertItem( locale->translate("&How to play"), this, SLOT(HowToPlay()) );

  menu = new KMenuBar( this );
  CHECK_PTR( menu );
  menu->insertItem( locale->translate("&File"), file );
  menu->insertItem( locale->translate("&Edit"), edit );
  menu->insertItem( locale->translate("&Skill"), skill );
  menu->insertItem( locale->translate("&Help"), help );

  pGame = new Ttt3dGame();

  pNormalPen = new QPen();
  pBoldPen   = new QPen( red, 2 );

  StatusBar = new KStatusBar(this);
  StatusBar->insertItem(locale->translate("Your move"), SB_MESSAGE );

  setStatusBar( StatusBar );
  setMenu( menu );

  setMinimumSize( 400, 600 );
}

Kttt3d::~Kttt3d()
{
  delete pNormalPen;
  delete pBoldPen;
  delete pGame;
}

void Kttt3d::NewGame()
{
  delete pGame;
  pGame = new Ttt3dGame( Skill );
  edit->setItemEnabled( RevertID, TRUE );
  edit->setItemEnabled( undoID, FALSE );
  StatusBar->changeItem(locale->translate("Your move"), SB_MESSAGE );
  repaint( TRUE );
}

void Kttt3d::undo()
{
  if( !pGame->IfThinking() ) {
    if( pGame->Undo() ) {
      repaint( TRUE );

      if( !pGame->GetMoveNumber() ) {
        edit->setItemEnabled( undoID, FALSE );
      }

      edit->setItemEnabled( RevertID, TRUE );

      StatusBar->changeItem(locale->translate("Your move"), SB_MESSAGE );
    }
  }
}

void Kttt3d::revert()
{
  if( !pGame->IfGameOver() ) {
    if( !pGame->IfThinking() ) {
      pGame->RevertBoard();

      GetComputerMove( TRUE );

      edit->setItemEnabled( undoID, TRUE );
    }
  }
}

void Kttt3d::UpdateSkillCheched( int Num )
{
  int i;

  for( i = 0; i < MAX_SKILL_ID; i++ ) {
    if( i == Num ) {
      skill->setItemChecked( SkillIDArr[ i ], TRUE );
    }
    else {
      skill->setItemChecked( SkillIDArr[ i ], FALSE );
    }
  }
}

void Kttt3d::easy()
{
  if( Skill != Ttt3dGame::BEGINNER ) {
    UpdateSkillCheched( 0 );
    pGame->SetSkill( Ttt3dGame::BEGINNER );
  }
}

void Kttt3d::normal()
{
  if( Skill != Ttt3dGame::NOVICE ) {
    UpdateSkillCheched( 1 );
    pGame->SetSkill( Ttt3dGame::NOVICE );
  }
}

void Kttt3d::difficult()
{
  if( Skill != Ttt3dGame::MASTER ) {
    UpdateSkillCheched( 2 );
    pGame->SetSkill( Ttt3dGame::MASTER );
  }
}

void Kttt3d::about()
{
    QMessageBox::message( "About kttt3d",
			"kttt3d by\n
			 Andrei Plekhanov <andyp@newtech.ru>\n
			 This program is covered by the GPL." );
}

void Kttt3d::HowToPlay()
{
  QMessageBox::message( "Help", "A three-dimensional tic-tac-toe is played on a 4x4x4 cube,
thus consisting of 64 small cubes.
Players move alternately by occupying any empty cube.
The game ends as soon as one of the players has
occupied four consecutive cubes in a straight line
( either in one, two or three dimensons )." );
}

void Kttt3d::GetComputerMove( int Update )
{
  int Result;

  StatusBar->changeItem(locale->translate("Thinking..."), SB_MESSAGE );

  repaint( Update );

	QApplication::setOverrideCursor( waitCursor );

  Result = pGame->ComputerMove();

  QApplication::restoreOverrideCursor();

  if( pGame->IfGameOver() ) {
    edit->setItemEnabled( RevertID, FALSE );
  }

  repaint( Update );

  switch( Result ) {
    case Ttt3dGame::CROSS_WIN:
      StatusBar->changeItem(locale->translate("Your lost"), SB_MESSAGE );
    break;
    case Ttt3dGame::NOL_WIN:
      StatusBar->changeItem(locale->translate("Your win"), SB_MESSAGE );
    break;
    case Ttt3dGame::CONTINUE_PLAY:
       StatusBar->changeItem(locale->translate("Your move"), SB_MESSAGE );
    break;
    case Ttt3dGame::DRAW_GAME:
       StatusBar->changeItem(locale->translate("Draw game"), SB_MESSAGE );
    break;
    case Ttt3dGame::CROSS_ATTACK:
       StatusBar->changeItem(locale->translate("You have only one move"), SB_MESSAGE );
    break;
    case Ttt3dGame::NOL_ATTACK:
      StatusBar->changeItem(locale->translate("Computer has only one move. You move"), SB_MESSAGE );
    break;
    case Ttt3dGame::CROSS_CHAIN:
      StatusBar->changeItem(locale->translate("Computer know the win"), SB_MESSAGE );
    break;
  }
}

void Kttt3d::SetMove( int Move )
{
  if( !pGame->IfGameOver() ) {
    if( !pGame->IfThinking() ) {
      if( pGame->IfEmpty( Move ) ) {
        pGame->PutNol( Move );

        pGame->SetHistory( Move, Ttt3dGame::HUMAN_MOVE );

//QMessageBox::warning( this, "Save failed", "Error saving file" );

        GetComputerMove( TRUE );

        edit->setItemEnabled( undoID, TRUE );
      }
    }
  }
}

void Kttt3d::ShowOnePos( QPainter * paint,
                         int X_LeftBottom, int Y_LeftBottom,
                         int X_RightBottom, int Y_RightBottom,
                         int X_LeftTop, int Y_LeftTop,
                         int X_RightTop, int Y_RightTop,
                         int Pos ) const
{
  int         Cross;
  int         BoldPenFlag;
  const Ttt3dMove * pMove;

  if( !pGame->IfEmpty( Pos ) ) {
    Cross = 1;

    if( pGame->GetFirst() == Ttt3dGame::HUMAN_FIRST ) {
      if( pGame->IfCross( Pos ) ) {
        Cross = 0;
      }
    }
    else {
      if( pGame->IfNol( Pos ) ) {
        Cross = 0;
      }
    }

    pMove = pGame->GetHistoryMove( pGame->GetMoveNumber() - 1 );

    BoldPenFlag = 0;

    if( pGame->IfGameOver() ) {
      if( pGame->WinLine( Pos ) ) {
        BoldPenFlag = 1;
      }
    }
    else {
      if( pMove != NULL ) {
        if( Pos == pMove->GetMove() ) {
          BoldPenFlag = 1;
        }
      }
    }

    if( BoldPenFlag ) {
      paint->setPen( *pBoldPen );
    }

    if( Cross ) {
      paint->drawLine( X_LeftBottom, Y_LeftBottom, X_RightTop, Y_RightTop );
      paint->drawLine( X_LeftTop, Y_LeftTop, X_RightBottom, Y_RightBottom );
    }
    else {
      paint->drawEllipse( X_LeftBottom + ( X_LeftTop - X_LeftBottom ) / 1.5,
                          Y_LeftTop,
                          X_RightBottom + ( X_RightTop - X_RightBottom ) / 2 -
                          X_LeftBottom - ( X_LeftTop - X_LeftBottom ) / 1.5,
                          Y_RightBottom - Y_RightTop );
    }

    if( BoldPenFlag ) {
      paint->setPen( *pNormalPen );
    }
  }
}

void Kttt3d::PaintLayer( QPainter * paint, int Number ) const
{
  int i;
  int j;
  int XDelta;
  int YDelta;

  XDelta = width() / ( 4 + 4 / X_OFF + 2 * X_SPACE );
  YDelta = height() / ( 4 * 4 + 3 * GAP + Y_SPACE1 + Y_SPACE2 );

  for( i = 0; i < 5; i++ ) {
    paint->drawLine( X_SPACE * XDelta + ( ( 2 - i ) * XDelta ) / X_OFF,
                     ( Y_SPACE1 + ( 4 + GAP ) * Number + i ) * YDelta,
                     ( X_SPACE + 4 ) * XDelta + ( ( 2 - i ) * XDelta ) / X_OFF,
                     ( Y_SPACE1 + ( 4 + GAP ) * Number + i ) * YDelta );

    paint->drawLine( ( X_SPACE + i ) * XDelta + ( 2 * XDelta ) / X_OFF,
                     ( Y_SPACE1 + ( 4 + GAP ) * Number ) * YDelta,
                     ( X_SPACE + i ) * XDelta - ( 2 * XDelta ) / X_OFF,
                     ( Y_SPACE1 + ( 4 + GAP ) * Number + 4 ) * YDelta );
  }

  for( i = 0; i < 4; i++ ) {
    for( j = 0; j < 4; j++ ) {
      ShowOnePos( paint, ( X_SPACE + j ) * XDelta + ( ( 1 - i ) * XDelta ) / X_OFF,
                         ( Y_SPACE1 + ( 4 + GAP ) * Number + i + 1 ) * YDelta,
                         ( X_SPACE + j + 1 ) * XDelta + ( ( 1 - i ) * XDelta ) / X_OFF,
                         ( Y_SPACE1 + ( 4 + GAP ) * Number + i + 1 ) * YDelta,
                         ( X_SPACE + j ) * XDelta + ( ( 2 - i ) * XDelta ) / X_OFF,
                         ( Y_SPACE1 + ( 4 + GAP ) * Number + i ) * YDelta,
                         ( X_SPACE + j + 1 ) * XDelta + ( ( 2 - i ) * XDelta ) / X_OFF,
                         ( Y_SPACE1 + ( 4 + GAP ) * Number + i ) * YDelta,
                         16 * Number + 4 * i + j );
    }
  }
}

float Kttt3d::LinePos( float X1, float X2, float Y1,
                       float Y2, float X_Pos, float Y_Pos ) const
{
  return( ( X_Pos - X1 ) / ( X1 - X2 ) +
          ( Y_Pos - Y1 ) / ( Y2 - Y1 ) );
}

int Kttt3d::CheckOnePos( int X_LeftTop, int X_RightTop,
                         int X_LeftBottom, int X_RightBottom,
                         int Y_Bottom, int Y_Top,
                         int X_Pos, int Y_Pos ) const
{
  if( Y_Pos > Y_Bottom ) {
    if( Y_Pos < Y_Top ) {
      if( LinePos( (float)X_LeftBottom, (float)X_LeftTop, (float)Y_Bottom,
                   (float)Y_Top, (float)X_Pos, (float)Y_Pos ) > 0.0 ) {
        if( LinePos( (float)X_RightBottom, (float)X_RightTop, (float)Y_Bottom,
                     (float)Y_Top, (float)X_Pos, (float)Y_Pos ) < 0.0 ) {
          return( 1 );
        }
      }
    }
  }

  return( 0 );
}

int Kttt3d::GetMove( int X_Pos, int Y_Pos ) const
{
  int i;
  int j;
  int k;
  int XDelta;
  int YDelta;

  XDelta = width() / ( 4 + 4 / X_OFF + 2 * X_SPACE );
  YDelta = height() / ( 4 * 4 + 3 * GAP + Y_SPACE1 + Y_SPACE2 );

  for( i = 0; i < 4; i++ ) {
    for( j = 0; j < 4; j++ ) {
      for( k = 0; k < 4; k++ ) {
        if( CheckOnePos( ( X_SPACE + k ) * XDelta + ( ( 1 - j ) * XDelta ) / X_OFF,
                         ( X_SPACE + k + 1 ) * XDelta + ( ( 1 - j ) * XDelta ) / X_OFF,
                         ( X_SPACE + k ) * XDelta + ( ( 2 - j ) * XDelta ) / X_OFF,
                         ( X_SPACE + k + 1 ) * XDelta + ( ( 2 - j ) * XDelta ) / X_OFF,
                         ( Y_SPACE1 + ( 4 + GAP ) * i + j ) * YDelta,
                         ( Y_SPACE1 + ( 4 + GAP ) * i + j + 1 ) * YDelta,
                         X_Pos, Y_Pos ) ) {
          return( 16 * i + 4 * j + k );
        }
      }
    }
  }

  return( -1 );
}

void Kttt3d::mousePressEvent( QMouseEvent * e )
{
  int Move;

  if( e->button() == LeftButton ) {
    Move = GetMove( e->x(), e->y() );

    if( Move >= 0 ) {
      SetMove( Move );
    }
  }
}

void Kttt3d::ShowHistory( QPainter * paint ) const
{
  int               i;
  const Ttt3dMove * pMove;
  QString           s;
  int               XDelta;
  int               YDelta;

  XDelta = X_SHIFT * width() / X_KOEF;;
  YDelta = Y_SHIFT * height() / Y_KOEF;

  for( i = 0; i < pGame->GetMoveNumber(); i++ ) {
    pMove = pGame->GetHistoryMove( i );

    if( pMove->GetType() == Ttt3dGame::COMPUTER_MOVE ) {
      s.sprintf("Computer %d %d %d", pMove->GetMove() / 16 + 1,
                                     ( pMove->GetMove() / 4 ) % 4 + 1,
                                     pMove->GetMove() % 4 + 1 );
    }
    else {
      s.sprintf("Human    %d %d %d", pMove->GetMove() / 16 + 1,
                                     ( pMove->GetMove() / 4 ) % 4 + 1,
                                     pMove->GetMove() % 4 + 1 );
    }

    paint->drawText( XDelta, YDelta + 20 * i, s );
  }
}

void Kttt3d::paintEvent( QPaintEvent * )
{
  int         i;
  QPainter    paint( this );

  for( i = 0; i < 4; i++ ) {
    PaintLayer( &paint, i );
  }

  ShowHistory( &paint );
}