/***************************************************************************
                          ttt3dgame.cpp  -  description
                             -------------------
    begin                : Thu Apr 27 2000
    copyright            : (C) 2000 by Andrei Plekhanov
    email                : andyp@newtech.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "ttt3dgame.h"
#include <stdio.h>

const int MAX_DEPTH_CROSS = 4;

Ttt3dGame::Ttt3dGame( int aSkill ) : ticboard()
{
  Skill      = aSkill;
  MoveNumber = 0;
  Thinking   = 0;
  GameOver   = 0;
  Chain      = 0;
  First      = HUMAN_FIRST;

  ClearWinMove();
  History = new Ttt3dMove * [ 64 ];
}

Ttt3dGame::~Ttt3dGame( void )
{
  int         i;
  Ttt3dMove * pMove;

  for( i = 0; i < MoveNumber; i++ ) {
    pMove = History[ i ];
    delete pMove;
  }

  delete History;
}

void Ttt3dGame::ClearWinMove( void )
{
  int i;

  for( i = 0; i < 64; i++ ) {
    WinMove[ i ] = -1;
  }
}

int Ttt3dGame::SetSkill( int aSkill )
{
  switch( aSkill ) {
    case BEGINNER:
    case NOVICE:
    case MASTER:
      Skill = aSkill;
    return( 0 );
  }

  return( 1 );
}

void Ttt3dGame::SetHistory( int Move, int Type )
{
  Ttt3dMove * pMove;

  pMove = new Ttt3dMove( Move, Type );

  History[ MoveNumber ] = pMove;

  MoveNumber++;
}

void Ttt3dGame::DelHistory( void )
{
  Ttt3dMove * pMove;

  MoveNumber--;

  pMove = History[ MoveNumber ];

  delete pMove;
}

int Ttt3dGame::IfCross( int Pos ) const
{
  int Val;

  Val = GetBoard( Pos );

  if( Val == CROSS ) {
    return( 1 );
  }

  return( 0 );
}

int Ttt3dGame::IfNol( int Pos ) const
{
  int Val;

  Val = GetBoard( Pos );

  if( Val == NOL ) {
    return( 1 );
  }

  return( 0 );
}

int Ttt3dGame::IfEmpty( int Pos ) const
{
  int Val;

  Val = GetBoard( Pos );

  if( Val == EMPTY ) {
    return( 1 );
  }

  return( 0 );
}

int Ttt3dGame::CheckDraw( void ) const
{
  int i;

  for( i = 0; i < 64; i++ ) {
    if( GetBoard( i ) == EMPTY ) {
      return( 0 );
    }
  }

  return( 1 );
}

void Ttt3dGame::RevertBoard( void )
{
  int i;

  for( i = 0; i < 64; i++ ) {
    if( GetBoard( i ) == CROSS ) {
      DelCross( i );
      PutNol( i );
    }
    else {
      if( GetBoard( i ) == NOL ) {
        DelNol( i );
        PutCross( i );
      }
    }
  }

  if( First == HUMAN_FIRST ) {
    First = COMPUTER_FIRST ;
  }
  else {
    if( First == COMPUTER_FIRST ) {
      First = HUMAN_FIRST;
    }
  }

  Chain = 0;
}

int Ttt3dGame::ComputerMove( void )
{
  int         Move;
  int         ResLine;

  Thinking = 1;

  ResLine = CheckAllLines( 4 * NOL );

  if( ResLine >= 0 ) {
    GameOver = 1;
    Thinking = 0;
    return( NOL_WIN );
  }

  if( CheckDraw() ) {
    GameOver = 1;
    Thinking = 0;
    return( DRAW_GAME );
  }

  ResLine = CheckAllLines( 3 * CROSS );

  if( ResLine >= 0 ) {
    Move = TestLine( ResLine );

    SetHistory( Move, COMPUTER_MOVE );

    PutCross( Move );
    GameOver = 1;
    Thinking = 0;
    return( CROSS_WIN );
  }

  ResLine = CheckAllLines( 3 * NOL );

  if( ResLine >= 0 ) {
    Move = TestLine( ResLine );

    SetHistory( Move, COMPUTER_MOVE );

    PutCross( Move );

    if( Chain ) {
      Chain = Chain + 2;
    }

    Thinking = 0;
    return( NOL_ATTACK );
  }

  if( Chain ) {
    Move = WinMove[ Chain ];
    Chain = Chain + 2;
  }
  else {
    ClearWinMove();

    if( FindFastCrossW( MAX_DEPTH_CROSS + GetSkill(), WinMove ) == 1 ) {
      Move = WinMove[ 0 ];
      Chain = 2;
    }
    else {
      Move = GetComputerMove( GetSkill() );

      if( Move < 0 ) {
        GameOver = 1;
      }
    }
  }

  SetHistory( Move, COMPUTER_MOVE );

  PutCross( Move );
  Thinking = 0;

  if( CheckDraw() ) {
    GameOver = 1;
    Thinking = 0;
    return( DRAW_GAME );
  }

  if( Chain ) {
    return( CROSS_CHAIN );
  }

  ResLine = CheckAllLines( 3 * CROSS );

  if( ResLine >= 0 ) {
    return( CROSS_ATTACK );
  }

  return( CONTINUE_PLAY );
}

void Ttt3dGame::UndoOneMove( void )
{
  int LastMove;

  MoveNumber--;

  LastMove = History[ MoveNumber ]->GetMove();

  if( GetBoard( LastMove ) == CROSS ) {
    DelCross( LastMove );
  }
  else {
    if( GetBoard( LastMove ) == NOL ) {
      DelNol( LastMove );
    }
  }

  if( Chain ) {
    Chain--;
  }

  if( GameOver ) {
    GameOver = 0;
    Chain = 0;
  }
}

int Ttt3dGame::Undo( void )
{
  if( MoveNumber > 1 ) {
    if( History[ MoveNumber - 2 ]->GetType() !=
        History[ MoveNumber - 1 ]->GetType() ) {
      UndoOneMove();
    }
    else {
      RevertBoard();
    }

    UndoOneMove();

    return( 1 );
  }
  else {
    if( MoveNumber > 0 ) {
      UndoOneMove();
      RevertBoard();
      return( 1 );
    }
  }

  return( 0 );
}

const Ttt3dMove * Ttt3dGame::GetHistoryMove( int Num ) const
{
  if( MoveNumber ) {
    if( Num < MoveNumber ) {
      return( History[ Num ] );
    }
  }

  return( NULL );
}

int Ttt3dGame::WinLine( int Pos ) const
{
  int LineNumber;

  LineNumber = -1;

  if( GetBoard( Pos ) == CROSS ) {
    LineNumber = Check( Pos, 4 * CROSS );
  }
  else {
    if( GetBoard( Pos ) == NOL ) {
      LineNumber = Check( Pos, 4 * NOL );
    }
  }

  if( LineNumber >= 0 ) {
    return( 1 );
  }

  return( 0 );
}
